// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <memory>
#include <mutex>
#include "xmacpp/Frame.h"
#include "InternalSessionFactory.h"
#include "BufferPool.h"
#include "vpi_pic.h"

namespace xma {
//////////////////////////////////////////////////////////////////////////////
class InternalFrameProperties final : public FrameProperties {
public:
  InternalFrameProperties(IntSessionFactoryPtr sf, LogPtr log, enum XmaFormatType format, enum XmaFormatType sw_format, int32_t width, int32_t height, uint32_t flags);
  InternalFrameProperties(IntSessionFactoryPtr sf, LogPtr log, XmaFrameProperties* props);
  virtual ~InternalFrameProperties() = default;

  XmaFrameProperties GetFrameProperties() const;

  static char*   GetPixFmtName(XmaFormatType fmt);
  static int32_t GetBufferSize(const FramePropertiesPtr props);

  virtual enum XmaFormatType GetFormat() const override;
  virtual enum XmaFormatType GetSwFormat() const override;
  virtual int32_t            GetWidth() const override;
  virtual int32_t            GetHeight() const override;
  virtual uint32_t           GetFlags() const override;
  virtual int32_t            GetNumPlanes() const override;
  virtual int32_t            GetPlaneHeight(int32_t plane) const override;
  virtual int32_t            GetPlaneStride(int32_t plane) const override;
  virtual int32_t            GetPlaneSize(int32_t plane) const override;

protected:
  XmaFrameProperties   m_props;
  IntSessionFactoryPtr m_sf;
  LogPtr               m_log;
};

using IntFramePropertiesPtr = std::shared_ptr<InternalFrameProperties>;

int32_t AddFrameToVpiFrm(IntSessionFactoryPtr sf, LogPtr log, FramePtr frame, VpiFrm* pic);
int32_t CopyFrame(LogPtr log, XmaFrame* dst, FramePtr src);

//////////////////////////////////////////////////////////////////////////////
class InternalFrame final : public Frame {
public:
  InternalFrame(IntSessionFactoryPtr sf, LogPtr log, FramePropertiesPtr props, bool dummy, buffer_t buffer, XmaFraction frame_rate = XMA_DEFAULT_FRAMERATE,
      int64_t pts = XMA_AV_NOPTS_VALUE, bool is_idr = false);
  InternalFrame(IntSessionFactoryPtr sf, LogPtr log, FramePropertiesPtr props, uint8_t** frame_data, std::function<void(uint8_t**)> free_callback = nullptr,
      XmaFraction frame_rate = XMA_DEFAULT_FRAMERATE, int64_t pts = XMA_AV_NOPTS_VALUE, bool is_idr = false);
  InternalFrame(
      IntSessionFactoryPtr sf, LogPtr log, FramePtr frame, XmaFraction frame_rate = XMA_DEFAULT_FRAMERATE, int64_t pts = XMA_AV_NOPTS_VALUE, bool is_idr = false);
  InternalFrame(IntSessionFactoryPtr sf, LogPtr log, VpiFrm* pic, XmaFraction frame_rate = XMA_DEFAULT_FRAMERATE, XmaFormatType format = XMA_NONE_FMT_TYPE,
      XmaFormatType sw_format = XMA_NONE_FMT_TYPE, int32_t width = 0, int32_t height = 0);
  virtual ~InternalFrame();

  XmaFrame*      GetFrame();
  VpiBufferPool* GetBufferPool();
  void           ExtendBufferPool(uint32_t frames);
  void           SetSelf();
  void           SetFrozen(bool sidedata_only = false);
  bool           IsFrozen() const;
  void           CopySideData(FramePtr src);
  void           CopyUserSideData(FramePtr src);
  void           CopySideDataMetadata(FramePtr src);
  SideDataPtr    GetSideDataAtIndex(uint32_t index);

  virtual XmaBufferType      GetBufferType() const override;
  virtual void*              GetBuffer(int32_t plane) const override;
  virtual uint8_t*           GetHostBuffer(int32_t plane) const override;
  virtual void*              GetDeviceBuffer(int32_t plane) const override;
  virtual FramePropertiesPtr GetProps() const override;
  virtual XmaFraction        GetFrameRate() const override;
  virtual int64_t            GetPts() const override;
  virtual int                IsIdr() const override;
  virtual void               SetFrameRate(XmaFraction frame_rate) override;
  virtual void               SetPts(int64_t pts) override;
  virtual void               SetIdr(bool is_idr) override;

  virtual int32_t     IncRef() override;
  virtual int32_t     DecRef() override;
  virtual int32_t     GetRefCount() override;
  virtual int32_t     AddSideData(SideDataPtr side_data) override;
  virtual SideDataPtr GetSideData(enum XmaFrameSideDataType type) override;
  virtual int32_t     RemoveSideData(SideDataPtr side_data) override;
  virtual int32_t     RemoveSideDataType(enum XmaFrameSideDataType type) override;
  virtual void        ClearAllSideData() override;
  virtual SideDataPtr GetFirstSideData() override;
  virtual SideDataPtr GetNextSideData(SideDataPtr side_data) override;
  virtual SideDataPtr GetNextSideDataOfSameType(SideDataPtr side_data) override;

  friend int32_t AddFrameToVpiFrm(IntSessionFactoryPtr sf, LogPtr log, FramePtr frame, VpiFrm* pic);
  friend int32_t CopyFrame(LogPtr log, XmaFrame* dst, FramePtr src);

protected:
  void CopySideData(VpiFrm* src);

protected:
  std::unique_ptr<XmaFrame> m_frame;
  FramePropertiesPtr        m_props;
  IntSessionFactoryPtr      m_sf;
  LogPtr                    m_log;
};

using IntFramePtr = std::shared_ptr<InternalFrame>;

VpiBufferPool* GetVpiPool(LogPtr log, FramePtr frame);

//////////////////////////////////////////////////////////////////////////////
#define FRAME_DUMMY_VALUE 0xABCDEF
class PrivateFrameInfo {
public:
  PrivateFrameInfo(IntSessionFactoryPtr sf, LogPtr log, buffer_t buffer = nullptr);
  ~PrivateFrameInfo();

  bool                  IsValid() const;
  std::recursive_mutex& GetMutex();
  VpiBufferPool*        GetBufferPool();
  void                  SetSelf(IntFramePtr self);
  IntFramePtr           GetSelf() const;
  void                  SetParent(FramePtr parent);
  int32_t               AddVpiFrm(VpiFrm* pic);
  VpiFrm*               GetVpiFrm() const;
  int32_t               RemoveVpiFrm();
  void                  SetCallback(std::function<void(uint8_t**)> free_callback);
  void                  CallCallback(uint8_t** frame_data);
  void                  SetFrozen();
  void                  Unfreeze(); // Only called when destroying frame
  bool                  IsFrozen() const;
  IntSessionFactoryPtr  GetSessionFactory();
  LogPtr                GetLog();

private:
  using vpi_frm_t = std::unique_ptr<VpiFrm, std::function<void(VpiFrm*)>>;
  mutable std::recursive_mutex   m_mutex;
  uint32_t                       m_dummy = FRAME_DUMMY_VALUE;
  buffer_t                       m_buffer;
  IntFramePtr                    m_self;   // C implementation uses this
  FramePtr                       m_parent; // Used for cloning another frame
  vpi_frm_t                      m_pic;    // Frame was made from VpiFrm
  std::function<void(uint8_t**)> m_free_callback = nullptr;
  bool                           m_frozen        = false;
  IntSessionFactoryPtr           m_sf;
  LogPtr                         m_log;
};

#define LOCK_XMA_FRAME(log, func, module, frame) \
  if (!frame) { \
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, module, "%s() XmaFrame is not valid!", func); \
  } \
  xma::PrivateFrameInfo* private_frame_info = static_cast<xma::PrivateFrameInfo*>(frame->private_do_not_touch); \
  if ((!private_frame_info) || (!private_frame_info->IsValid())) { \
    THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, module, "%s() Private info is not valid!", func); \
  } \
  std::unique_lock<std::recursive_mutex> guard(private_frame_info->GetMutex());

//////////////////////////////////////////////////////////////////////////////
#define VPIFRM_DUMMY_VALUE 0x123456
typedef void (*VpePicFree)(void* opaque, void* data, void* buf);

struct VpiFrmInfo {
  FramePtr   frame;
  VpiBuffer* buf;
};

using vpi_frame_info_t = std::unique_ptr<VpiFrmInfo>;

class PrivateVpiFrmInfo {
public:
  PrivateVpiFrmInfo(IntSessionFactoryPtr sf, LogPtr log);
  ~PrivateVpiFrmInfo();

  bool                 IsValid() const;
  int32_t              AddFrame(FramePtr frame, VpiFrm* pic, int32_t ch);
  vpi_frame_info_t     GetFrame(XmaFormatType format, XmaFormatType sw_format, int32_t width, int32_t height);
  int32_t              GetFrameCount() const;
  int32_t              SetFreeBuf(VpePicFree func_free);
  VpePicFree           GetFreeBuf() const;
  int32_t              SetFreeData(void* data);
  void*                GetFreeData() const;
  int32_t              SetPool(VpiBufferPool* pool);
  VpiBufferPool*       GetPool() const;
  IntSessionFactoryPtr GetSessionFactory();
  LogPtr               GetLog();

private:
  using vpi_frm_t                     = std::unique_ptr<VpiFrm, std::function<void(VpiFrm*)>>;
  uint32_t                    m_dummy = VPIFRM_DUMMY_VALUE;
  std::list<vpi_frame_info_t> m_frames;              // AddFrameToVpiFrm - list of frames
  VpePicFree                  m_func_free = nullptr; // AddFrameToVpiFrm - old VpePicFree
  void*                       m_free_data = nullptr; // used by encoder in VpiFrm to keep data needed to be freed
  VpiBufferPool*              m_pool      = nullptr; // AddFrameToVpiFrm - old VpiBufferPool
  IntSessionFactoryPtr        m_sf;
  LogPtr                      m_log;
};

void InitVpiFrm(VpiFrm* pic);

}; // Namespace xma
