// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "xlnx_ml_app.h"

int signal_caught;

/**
 * xlnx_ml_signal_handler: Signal handler function
 * @param _signum: Signal number
 */
static void xlnx_ml_signal_handler(int32_t signum)
{
    switch (signum) {
        case SIGHUP:
        case SIGINT:
        case SIGQUIT:
        case SIGABRT:
        case SIGTERM:
            signal_caught = 1;
            break;
    }
}

/**
 * xlnx_ml_set_signal_handler: Signal handler initialization.
 * @return XMA_APP_SUCCESS or APP_FAILURE
 */
static int xlnx_ml_set_signal_handler()
{
    signal_caught = 0;
    struct sigaction action;
    action.sa_handler = xlnx_ml_signal_handler;
    sigemptyset(&action.sa_mask);
    action.sa_flags = 0;
    sigaction(SIGHUP, &action, NULL);
    sigaction(SIGINT, &action, NULL);
    sigaction(SIGQUIT, &action, NULL);
    sigaction(SIGABRT, &action, NULL);
    sigaction(SIGTERM, &action, NULL);
    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_cleanup_ctx: Cleanup the context - free resources,
 * close files.
 *
 * @param ctx: The ml context
 */
void xlnx_ml_cleanup_ctx(XlnxMLAppCtx* ctx)
{
    if (ctx->in_fp) {
        fclose(ctx->in_fp);
    }
    if (ctx->out_fp) {
        fclose(ctx->out_fp);
    }
    if (ctx->out_tensor_fp) {
        fclose(ctx->out_tensor_fp);
    }
    xlnx_ml_cleanup_ml_ctx(&ctx->ml_ctx);
}

/**
 * xlnx_ml_set_output_params: Open output files and set pix fmt
 * @param arguments: The arguments struct containing commandline info
 * @param ctx: The ml app ctx
 * @return XMA_APP_SUCCESS or XMA_APP_ERROR
 */
static int32_t xlnx_ml_set_output_params(XlnxMLArguments arguments,
                                           XlnxMLAppCtx* ctx)
{
    ctx->out_fp = fopen(arguments.output_file, "wb");
    if (!ctx->out_fp) {
        printf("Unable to open output file %s\n", arguments.output_file);
        return XMA_APP_ERROR;
    }
    ctx->out_tensor_fp = fopen("output_tensor_from_xma.dat", "wb");
    if (!ctx->out_tensor_fp) {
        printf("Unable to open output tensor file\n");
        return XMA_APP_ERROR;
    }
    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_create_xma_param_ctx: Create the param ctx for xma
 * props.
 *
 * @param arguments: The arguments struct containing commandline info
 * @param ctx: The ml context
 * @return XMA_APP_SUCCESS on success XMA_APP_ERROR on error
 */
static int32_t xlnx_ml_create_xma_param_ctx(XlnxMLArguments arguments,
                                              XlnxMLProps*  param_ctx)
{
    // Setup custom xma plugin params
    param_ctx->device_id       = arguments.device_id;
    param_ctx->log_level       = arguments.log_level;

    param_ctx->input_pix_fmt   = arguments.input_pix_fmt;
    param_ctx->input_width     = arguments.input_width;
    param_ctx->input_height    = arguments.input_height;
  
    param_ctx->input_fps_num   = arguments.fps_num;
    param_ctx->input_fps_den   = arguments.fps_den;

    param_ctx->model_name      = arguments.model_name;
    param_ctx->model_arg       = arguments.model_arg;
    param_ctx->ml_perf         = arguments.ml_perf;
    param_ctx->ml_tensor_dump  = arguments.ml_tensor_dump;
    param_ctx->ml_inf_repeat   = arguments.ml_inf_repeat;

    param_ctx->num_outputs = 1;
    param_ctx->num_inputs = 1;

    if (param_ctx->input_pix_fmt == XMA_YUV420P10LE_FMT_TYPE) {
        param_ctx->input_stride = 2 * arguments.input_width;
    } else {
        param_ctx->input_stride = arguments.input_width;
    }

    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_create_context: Creates the context based on user 
 * arguments. It parses the header of the input file to get 
 * relevant codec info. This does not create the xma session. 
 * Nor does it initialize the device. 
 * @param arguments: The argument struct containing ml param, 
 * input, output file 
 * @param info
 * @param ctx: A pointer to the ml context
 * @return XMA_APP_SUCCESS on success
 */
static int32_t xlnx_ml_create_context(XlnxMLArguments arguments,
                                        XlnxMLAppCtx* ctx)
{
    ctx->num_frames_to_inference = arguments.num_frames;
    ctx->in_fp                   = fopen(arguments.input_file, "rb");
    if (!ctx->in_fp) {
        printf("Unable to open input file %s\n", arguments.input_file);
        return XMA_APP_ERROR;
    }
    ctx->in_pix_fmt = arguments.input_pix_fmt;
    xlnx_ml_create_xma_param_ctx(arguments, &ctx->ml_ctx.ml_params);
    // Open output files, set output pix fmts //
    if (xlnx_ml_set_output_params(arguments, ctx) != XMA_APP_SUCCESS) {
        return XMA_APP_ERROR;
    }
    if (xlnx_ml_create_ml_ctx(&ctx->ml_ctx) != XMA_APP_SUCCESS) {
        return XMA_APP_ERROR;
    }
    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_get_input_frame: Read an input frame into the xframe 
 * buffer 
 * @param ctx: A pointer to the ml context
 * @param in_fp: ctx->input_ctx.in_fp
 * @param xframe: ctx->input_ctx.xframe
 * @param props: ctx->ml_xma_props
 * @return XMA_APP_SUCCESS on success XMA_APP_ERROR on error
 */
static int32_t xlnx_ml_get_input_frame(XlnxMLAppCtx* ctx)
{
    int ret                      = XMA_APP_SUCCESS;
    ctx->ml_ctx.input_xframe[0]  = xma_frame_alloc(
        ctx->ml_ctx.handle, &ctx->ml_ctx.upload_fprops, false);
    if (!ctx->ml_ctx.input_xframe[0]) {
        xma_logmsg(ctx->ml_ctx.log, XMA_ERROR_LOG,
                   XLNX_ML_APP_MODULE,
                   "Failed to allocate xma frame for DMA to device!\n");
        return XMA_APP_ERROR;
    }
    XmaFrame*           input_xframe = ctx->ml_ctx.input_xframe[0];
    ret = xlnx_utils_read_frame(ctx->ml_ctx.handle, ctx->ml_ctx.log, input_xframe, ctx->in_fp);

    if (ret == XMA_APP_EOF) {
        xma_frame_free(input_xframe);
        ctx->ml_ctx.input_xframe[0] = NULL;
    }

    return ret;
}

/**
 * xlnx_ml_print_segment_performance: Print the performance 
 * since the last segment. 
 * @param ctx: The ml context
 */
static void xlnx_ml_print_segment_performance(XlnxMLAppCtx* ctx)
{
    double time_since_last_segment = xlnx_utils_get_segment_time(&ctx->timer);
    if (time_since_last_segment < 0.5) {
        return;
    }
    fprintf(stderr, "\rFrame=%5zu Total FPS=%.03f Current FPS=%.03f\r",
            ctx->num_frames_inferenced,
            (float)ctx->num_frames_inferenced /
                xlnx_utils_get_total_time(&ctx->timer),
            (ctx->num_frames_inferenced - ctx->timer.last_displayed_frame) /
                time_since_last_segment);
    fflush(stderr);
    ctx->timer.last_displayed_frame = ctx->num_frames_inferenced;
    xlnx_utils_set_segment_time(&ctx->timer);
}

/**
 * xlnx_ml_print_total_performance: Print the total performance 
 * since tracking began. 
 * @param ctx: The ml context
 */
static void xlnx_ml_print_total_performance(XlnxMLAppCtx* ctx)
{
    double realtime_taken = xlnx_utils_get_total_time(&ctx->timer);
    fprintf(stderr, "\nFrames Inferenced: %zu, Time Elapsed: %.03lf\r\n",
            ctx->num_frames_inferenced, realtime_taken);
    fprintf(stderr, "Real Time FPS: %.03lf\r\n",
            ctx->num_frames_inferenced / realtime_taken);
}

/**
 * Write the output frame for an output
 * @param ctx The ml app context
 * @param output_id The output index for the xframelist
 * @return XMA_APP_SUCCESS on success, XMA_APP_ERROR on error
 */
static int xlnx_ml_write_output_frame(XlnxMLAppCtx* ctx)
{
    int              ret;
    XlnxMLCtx*   ml_ctx    = &ctx->ml_ctx;
    XmaFrame*        output_xframe = ml_ctx->output_xframe_list[0];
    ret                            = xma_filter_session_send_frame(
        ml_ctx->download_sessions, output_xframe);
    if (ret != XMA_SUCCESS) {
        xma_logmsg(ml_ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "Upload filter failed to receive frame from filter!\n");
        return XMA_APP_ERROR;
    }
    output_xframe = xma_frame_alloc(
        ml_ctx->handle, &ml_ctx->download_fprops, true);
    if (!output_xframe) {
        xma_logmsg(ml_ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "Failed to allocate xma frame for DMA to host!\n");
        return XMA_APP_ERROR;
    }
    uint32_t counter = 1000000;
    do {
        ret = xma_filter_session_recv_frame(
            ml_ctx->download_sessions, output_xframe);
        if (ret == XMA_TRY_AGAIN) {
            usleep(100);
        }
    } while ((ret == XMA_TRY_AGAIN) && (--counter > 0));

    XmaFrameSideData *roi_sd = NULL;
    roi_sd = xma_frame_get_side_data(output_xframe, XMA_FRAME_SIDE_DATA_ML_ROI);
    if ((roi_sd)) {
        unsigned char *detected_roi = NULL;
        detected_roi = roi_sd->host;
        int retVal;
        retVal = fwrite(detected_roi, 1, roi_sd->size, ctx->out_tensor_fp);
        xma_side_data_free(roi_sd);
        if (!retVal) {
            xma_logmsg(ml_ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                       "Failed to write tensor dump returnVal: %d.\n", retVal);
            return XMA_APP_ERROR;
        }
    }

    if (ret == XMA_SUCCESS) {
        xlnx_utils_write_xframe(ml_ctx->handle, ml_ctx->log, output_xframe, ctx->out_fp);
    } else {
        xma_frame_free(output_xframe);
        ml_ctx->output_xframe_list[0] = NULL;
    }

    return XMA_APP_SUCCESS;
}

/**
 * Inference a frame
 * @param ctx The ml app context
 * @param session_id The id of the session; used to track mixrate
 * @return The result of sending the frame, or xma error if recv is not
 * successful.
 */
static int xlnx_ml_inference_frame(XlnxMLAppCtx* ctx)
{
    int ret;
    int send_rc=XMA_EOS, recv_rc;

    // send frame to ml //
    XlnxMLCtx* ml_ctx      = &ctx->ml_ctx;
    XmaFrame* input_xframe = ml_ctx->input_xframe[0];
    ret = xma_filter_session_send_frame(ml_ctx->upload_session, input_xframe);
    if (ret != XMA_SUCCESS) {
        xma_logmsg(ml_ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "Upload filter failed to send frame %d\n",
                   ml_ctx->num_frames_inferenced);
        return ret;
    }
    input_xframe =
        xma_frame_alloc(ml_ctx->handle, &ml_ctx->input_fprops, true);
    if (!input_xframe) {
        xma_logmsg(ml_ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "Failed to allocate xma frame for DMA to device!\n");
        return ret;
    }
    uint32_t counter = 1000000;
    do {
        ret =
            xma_filter_session_recv_frame(ml_ctx->upload_session, input_xframe);
        if (ret == XMA_TRY_AGAIN) {
            usleep(100);
        }
    } while ((ret == XMA_TRY_AGAIN) && (--counter > 0));
    if (ret == XMA_EOS) {
        xma_frame_free(input_xframe);
        input_xframe = NULL;
    } else if (ret != XMA_SUCCESS) {
        xma_logmsg(ml_ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                   "Upload filter failed to receive frame %d\n",
                   ml_ctx->num_frames_inferenced);
        xma_frame_free(input_xframe);
        input_xframe = NULL;
        return ret;
    }

    if (! input_xframe) {
        send_rc = xma_mio_filter_session_send_frame_list(ml_ctx->ml_session, NULL);
    } else {
        send_rc = xma_mio_filter_session_send_frame_list(ml_ctx->ml_session, &input_xframe);
    }

    // receive frame from ml //
    if ((send_rc == XMA_SUCCESS) || (send_rc == XMA_FLUSH_AGAIN)) {
        ml_ctx->output_xframe_list[0] = xma_frame_alloc(
            ctx->ml_ctx.handle, &ml_ctx->output_fprops, true);
        if (!ml_ctx->output_xframe_list[0]) {
            xma_logmsg(
                ml_ctx->log, XMA_ERROR_LOG, XLNX_ML_APP_MODULE,
                "Failed to allocate xma frame for ml receive!\n");
            return XMA_APP_ERROR;
        }
        counter = 1000000;
        do {
            recv_rc = xma_mio_filter_session_recv_frame_list(
                ml_ctx->ml_session, ml_ctx->output_xframe_list);
            if (recv_rc == XMA_TRY_AGAIN) {
                usleep(100);
            }
        } while ((recv_rc == XMA_TRY_AGAIN) && (--counter > 0));
        if (recv_rc == XMA_EOS) {
            xma_frame_free(ml_ctx->output_xframe_list[0]);
            ml_ctx->output_xframe_list[0] = NULL;
            send_rc = XMA_EOS;
        } else if (recv_rc != XMA_SUCCESS) {
            xma_logmsg(ml_ctx->log, XMA_ERROR_LOG,
                       XLNX_ML_APP_MODULE,
                       "Failed to receive frame list from XMA plugin\n");
            return XMA_ERROR;
        }
        xlnx_ml_write_output_frame(ctx);
        if (send_rc != XMA_EOS) {
            ctx->num_frames_inferenced++;
        }
        xlnx_ml_print_segment_performance(ctx);
    }
    return send_rc;
}

/**
 * xlnx_ml_inference_file: Perform inference on a file
 *
 * @param ctx: The ml context
 * @return XMA_APP_SUCCESS on success XMA_APP_ERROR on error
 */
static int xlnx_ml_inference_file(XlnxMLAppCtx* ctx)
{
    int send_rc = XMA_SUCCESS;
    xlnx_ml_set_signal_handler();
    xlnx_utils_set_non_blocking(1);
    xlnx_utils_start_tracking_time(&ctx->timer);
    while (send_rc != XMA_EOS && send_rc != XMA_ERROR) {
        if (xlnx_utils_was_q_pressed() || signal_caught ||
           ctx->num_frames_inferenced >= ctx->num_frames_to_inference) {
            fseek(ctx->in_fp, 0, SEEK_END);
        }
        xlnx_ml_get_input_frame(ctx);
        send_rc = xlnx_ml_inference_frame(ctx);
    }
    fprintf(stderr, "\n");
    xlnx_utils_set_non_blocking(0);
    return send_rc != XMA_APP_ERROR ? XMA_APP_SUCCESS : XMA_APP_ERROR;
}

/**
 * xlnx_ml_create_device_sessions: Create ml sessions
 * @param ctx: The ml context
 * @return XMA_APP_SUCCESS or APP_FAILURE
 */
static int32_t xlnx_ml_create_device_sessions(XlnxMLArguments* arguments,
                                              XlnxMLAppCtx*  ctx)
{
    XlnxMLCtx* ml_ctx = &ctx->ml_ctx;
    int ret = xma_log_init(arguments->log_level, arguments->log_location,
                           &ctx->ml_ctx.log, arguments->log_file);
    if (ret != XMA_SUCCESS) {
        fprintf(stderr, "XMA logging initialization failed\n");
        return XMA_APP_ERROR;
    }
    // load device
    
    if (xlnx_ml_device_init(ml_ctx) != XMA_APP_SUCCESS) {
        return XMA_APP_ERROR;
    }
    if (xlnx_ml_create_ml_sessions(ml_ctx) != XMA_APP_SUCCESS) {
        return XMA_APP_ERROR;
    }
    return XMA_APP_SUCCESS;
}

/**
 * main: Main routine
 * @param argc: Arguments count
 * @param argv: Argument array
 * @return XMA_APP_SUCCESS or APP_FAILURE
 */
int main(int argc, char* argv[])
{
    XlnxMLArguments arguments = xlnx_ml_get_arguments(argc, argv);
    XlnxMLAppCtx  ctx;
    memset(&ctx, 0, sizeof(XlnxMLAppCtx));
    if (xlnx_ml_create_context(arguments, &ctx) != XMA_APP_SUCCESS) {
        xlnx_ml_cleanup_ctx(&ctx);
        exit(XMA_APP_ERROR);
    }
    if (xlnx_ml_create_device_sessions(&arguments, &ctx) != XMA_APP_SUCCESS) {
        xlnx_ml_cleanup_ctx(&ctx);
        exit(XMA_APP_ERROR);
    }
    // run ml vip9000
    xlnx_ml_inference_file(&ctx);
    xlnx_ml_print_total_performance(&ctx);
    xlnx_ml_cleanup_ctx(&ctx);
    return XMA_APP_SUCCESS;
}
